"""Common GRPC backend client."""
import logging
import random
import string
from types import ModuleType
from typing import Any, Callable, Dict, Sequence

import grpc
from google.protobuf.message import Message
from google.protobuf.service import Service

from .decoder import GrpcDecoder
from .exceptions import ServiceNotFound
from .future import FutureCallWrapper

_LOGGER = logging.getLogger(__name__)


class GrpcClient(object):
    """GRPC backend client."""

    def __init__(self, netloc: str, decoder: GrpcDecoder, grpc_modules: Sequence[ModuleType],
                 credentials: grpc.ChannelCredentials = None):
        """
        Initialize GRPC backend client instance.

        Args:
            netloc: Network location of a gRPC server.
            decoder: Decoder instance.
            grpc_modules: List of generated _pb2_grpc modules with service stubs.
            credentials: Credentials for a secure channel connection. If None, insecure channel is used.
        """
        self.netloc = netloc
        self.credentials = credentials
        self.decoder = decoder
        self.grpc_modules = grpc_modules

        # Stubs cache
        self._stubs = {}  # type: Dict[str, Any]

        self.channel = self.connect()

    def connect(self) -> grpc.Channel:
        """Create channel to GPRC server."""
        if self.credentials is not None:
            return grpc.secure_channel(self.netloc, self.credentials)
        else:
            return grpc.insecure_channel(self.netloc)

    def get_stub(self, service: str) -> Service:
        """
        Return stub for the service.

        Stubs are cached, so for each service, there is only one stub created.
        """
        stub_name = '{}Stub'.format(service)

        if stub_name in self._stubs:
            return self._stubs[stub_name]

        for module in self.grpc_modules:
            if hasattr(module, stub_name):
                stub_cls = getattr(module, stub_name)
                stub = stub_cls(self.channel)
                self._stubs[stub_name] = stub
                return stub
        else:
            raise ServiceNotFound('Service {} could not be found'.format(stub_name))

    def _call(self, service: str, method: str, request: Message) -> Any:
        """
        Only call the grpc method.

        This method has four main steps:
            * calling grpc method
            * raising exception, if response contains any
            * decoding grpc result using decoder

        Args:
            service: Name of service, that is used to get service stub.
            method: Name of service method.
            request: Request message passed as an argument to service method.
        """
        stub = self.get_stub(service)
        grpc_method = getattr(stub, method).future

        try:
            response = grpc_method(request)
        except grpc.RpcError as error:
            if error.code() == grpc.StatusCode.UNAVAILABLE:
                # Connection probably failed. Try to reconnect and send the request again.
                _LOGGER.info('Connection to gRPC service %s failed. Trying to reconnect.', service)
                self.channel = self.connect()
                response = grpc_method(request)
            else:
                raise

        return response

    def call(
        self, service: str, method: str, request: Message, *,
        blocking: bool = True, modify_result: Callable[[Any], Any] = None
    ) -> Any:
        """
        Call grpc method.

        Args:
            service: Name of service, that is used to get service stub.
            method: Name of service method.
            request: Request message passed as an argument to service method.
            blocking: Determines whether to use blocking or non-blocking call.
            modify_result: Callable that processes decoded result and modifies it. Default is identity.
        """
        call_id = ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(4))
        _LOGGER.debug('[%s] %s(%s)', call_id, method, request)

        try:
            response = self._call(service, method, request)
        except grpc.RpcError as error:
            _LOGGER.info('[%s] %s failed with %s', call_id, method, error)
            raise

        _LOGGER.debug('[%s] %s returned %s', call_id, method, response)

        response_wrapper = FutureCallWrapper(response, decoder=self.decoder, modify_result=modify_result)
        if blocking:
            return response_wrapper.result()
        else:
            return response_wrapper
