"""Frgal decoder."""
from datetime import datetime
from typing import Any, Callable, Dict, Iterable, Mapping, Type

import pytz
from google.protobuf.message import Message
from google.protobuf.pyext._message import (MessageMapContainer, RepeatedCompositeContainer, RepeatedScalarContainer,
                                            ScalarMapContainer)
from google.protobuf.timestamp_pb2 import Timestamp

from .exceptions import DecoderError, FrgalException, UnknownError


class GrpcDecoder(object):
    """
    Generic grpc backend decoder.

    @ivar decoders: Dictionary of protobuf types and callables to decode them.
    @ivar exception_decoders: Dictionary of exception names and related exception classes.

    Decoding description:
        * Decoding is done recursively.
        * Primitive types are not decoded.
        * Empty string is replaced by None.
        * Repeated fields are decoded as arrays.
        * Map fields are decoded as dictionaries.
        * Messages are decoded as dictionaries.

    Default values handling:
        * Not present fields are assigned a default value.
        * Default value for repeated fields is [].
        * Default value for string field is None.
        * Default value for other primitive types fields are specified at
          https://developers.google.com/protocol-buffers/docs/proto3#default
        * Default value for any other field type is None.
    """

    def __init__(self):
        """Set common message type decoders."""
        self.decoders = {}  # type: Dict[Any, Callable]
        self.exception_decoders = {}  # type: Dict[str, Type[FrgalException]]

        for typ in (bool, int, str, float, bytes, type(None)):
            self.set_decoder(typ, lambda x: x)  # pragma: no cover

        # There are two different GRPC types for lists depending on their content
        self.set_decoder(RepeatedScalarContainer, self._decode_list)
        self.set_decoder(RepeatedCompositeContainer, self._decode_list)

        # There are two different GRPC types for maps depending on their content
        self.set_decoder(ScalarMapContainer, self._decode_dict)
        self.set_decoder(MessageMapContainer, self._decode_dict)

        self.set_decoder(Timestamp, self._decode_timestamp)
        self.set_decoder(Message, self._decode_struct)

    def set_decoder(self, object_type: Any, decode_func: Callable) -> None:
        """
        Set custom decoder.

        Args:
            object_type: Backend object type.
            decode_func (Callable): Callable getting backend value and returning decoded one.
        """
        self.decoders[object_type] = decode_func

    def set_exception_decoder(self, name: str, exception: Type[FrgalException]) -> None:
        """
        Set custom exception.

        Args:
            name: Backend exception name.
            exception: Python exception to be returned.
        """
        self.exception_decoders[name] = exception

    def decode(self, value: Any) -> Any:
        """
        Decode backend value.

        If there is custom decoder for the value type or its ancestors, it's used.
        Otherwise, exception is raised.
        """
        for cls in type(value).mro():
            if cls in self.decoders:
                return self.decoders[cls](value)

        raise DecoderError('Cannot decode unknown type {}'.format(type(value)))

    def decode_exception(self, value: Message) -> FrgalException:
        """
        Return appropriate exception type.

        Args:
            value: GRPC Exception Message
        """
        name = value.WhichOneof('Reasons')
        if name in self.exception_decoders:
            return self.exception_decoders[name]()
        else:
            return UnknownError(name)

    def _decode_list(self, value: Iterable) -> list:
        """Decode list."""
        return [self.decode(x) for x in value]

    def _decode_dict(self, value: Mapping) -> dict:
        """Decode dict."""
        # Keys are primitive types, so they don't need to be decoded.
        return {a: self.decode(b) for a, b in value.items()}

    def _decode_struct(self, value: Message) -> Any:
        """
        Decode GRPC Message object.

        Message without fields is decoded to None.
        Message with only one field is decoded to the decoding that field.
        Otherwise, Message is decoded to dictionary.
          * Keys of this dictionary are all that Message fields names, whether they're specified or not.
          * Specified fields are recursively decoded.
          * Non-specified fields are assigned default value.
        """
        possible_fields = value.DESCRIPTOR.fields
        fields = value.ListFields()
        if len(possible_fields) == 1:
            if len(fields) == 1:
                field, field_value = fields[0]
                return self.decode(field_value)
            else:
                return self._default_field_value(possible_fields[0])
        else:
            decoded = {}  # type: Dict[str, Any]
            for field in possible_fields:
                decoded[field.name] = self._default_field_value(field)

            for field, field_value in fields:
                decoded[field.name] = self.decode(field_value)
            return decoded

    @staticmethod
    def _default_field_value(field: Any) -> Any:
        """
        Return default value for given field.

          * For array field return [].
          * For string field return None.
          * For other primitive type field return its default value.
            See https://developers.google.com/protocol-buffers/docs/proto3#default
          * For other fields return None.
        """
        if field.label == field.LABEL_REPEATED:
            return []
        elif field.type == field.TYPE_STRING:
            return None
        elif field.message_type is None:
            return field.default_value
        else:
            return None

    @staticmethod
    def _decode_timestamp(timestamp: Timestamp) -> datetime:
        """Decode timestamp to aware datetime."""
        return datetime.fromtimestamp(timestamp.seconds + timestamp.nanos / 10**9, tz=pytz.utc)
