"""gRPC non-blocking calls utils."""
from typing import Any, Callable

import grpc
from google.protobuf.empty_pb2 import Empty

from .decoder import GrpcDecoder


class FutureCallWrapper:
    """gRPC future call wrapper."""

    def __init__(
        self,
        future_call: grpc.Future,
        decoder: GrpcDecoder,
        modify_result: Callable[[Any], Any] = None,
    ):
        """Initialize future call wrapper."""
        self._future_call = future_call
        self._decoder = decoder
        if modify_result is None:
            self._modify_result = lambda x: x
        else:
            self._modify_result = modify_result

    def result(self, timeout=None) -> Any:
        """Decode the result and possibly throw an exception."""
        response = self._future_call.result(timeout)

        # Assert the supported structure of messages
        assert ('exception' in response.DESCRIPTOR.fields_by_name or 'data' in response.DESCRIPTOR.fields_by_name
                or isinstance(response, Empty))

        # Decode exception first
        if 'exception' in response.DESCRIPTOR.fields_by_name and response.HasField('exception'):
            raise self._decoder.decode_exception(response.exception)  # type: ignore

        # Decode data otherwise
        if 'data' in response.DESCRIPTOR.fields_by_name and response.HasField('data'):
            return self._modify_result(self._decoder.decode(response.data))  # type: ignore

        # The response is empty
        return self._modify_result(None)

    def __getattr__(self, name):
        """Proxy all other future properties."""
        return getattr(self._future_call, name)
